# coding=utf-8
from celery.task import task as base_task, Task
import djcelery_transactions.transaction_signals
from django.db import transaction
from functools import partial
import threading


# Thread-local data (task queue).
_thread_data = threading.local()


def _get_task_queue():
    """Returns the calling thread's task queue."""
    return _thread_data.__dict__.setdefault("task_queue", [])


class PostTransactionTask(Task):
    """A task whose execution is delayed until after the current transaction.

    The task's fate depends on the outcome of the current transaction. If it's
    committed or no changes are made in the transaction block, the task is sent
    as normal. If it's rolled back, the task is discarded.

    If transactions aren't being managed when ``apply_asyc()`` is called (if
    you're in the Django shell, for example) or the ``after_transaction``
    keyword argument is ``False``, the task will be sent immediately.

    A replacement decorator is provided:

    .. code-block:: python

        from djcelery_transactions import task

        @task
        def example(pk):
            print "Hooray, the transaction has been committed!"
    """

    abstract = True

    @classmethod
    def original_apply_async(cls, *args, **kwargs):
        """Shortcut method to reach real implementation
        of celery.Task.apply_sync
        """
        return super(PostTransactionTask, cls).apply_async(*args, **kwargs)

    @classmethod
    def apply_async(cls, *args, **kwargs):
        # Delay the task unless the client requested otherwise or transactions
        # aren't being managed (i.e. the signal handlers won't send the task).
        if transaction.is_managed():
            if not transaction.is_dirty():
                # Always mark the transaction as dirty
                # because we push task in queue that must be fired or discarded
                if 'using' in kwargs:
                    transaction.set_dirty(using=kwargs['using'])
                else:
                    transaction.set_dirty()
            _get_task_queue().append((cls, args, kwargs))
        else:
            return cls.original_apply_async(*args, **kwargs)


def _discard_tasks(**kwargs):
    """Discards all delayed Celery tasks.

    Called after a transaction is rolled back."""
    _get_task_queue()[:] = []


def _send_tasks(**kwargs):
    """Sends all delayed Celery tasks.

    Called after a transaction is committed or we leave a transaction
    management block in which no changes were made (effectively a commit).
    """
    queue = _get_task_queue()
    while queue:
        cls, args, kwargs = queue.pop(0)
        cls.original_apply_async(*args, **kwargs)


# A replacement decorator.
task = partial(base_task, base=PostTransactionTask)

# Hook the signal handlers up.
transaction.signals.post_commit.connect(_send_tasks)
transaction.signals.post_rollback.connect(_discard_tasks)
transaction.signals.post_transaction_management.connect(_send_tasks)



# coding=utf-8
"""Adds signals to ``django.db.transaction``.

Signals are monkey patched into ``django.db.transaction``:

* ``post_commit``: sent after a transaction is committed. If no changes were
  made in the transaction block, nothing is committed and this won't be sent.
* ``post_rollback``: sent after a transaction is rolled back.
* ``post_transaction_management``: sent after leaving transaction management.
  This signal isn't  posted if a ``TransactionManagementError`` is raised.

.. code-block:: python

    import djcelery_transactions.transaction_signals


    def _post_commit(**kwargs):
        print "The transaction has been committed!"


    django.db.transaction.signals.post_commit.connect(_post_commit)

This code was inspired by Grégoire Cachet's implementation of similar
functionality, which can be found on GitHub: https://gist.github.com/247844

.. warning::

    This module must be imported before you attempt to use the signals.
"""
from functools import partial
import thread

from django.db import transaction
try:
    # Prior versions of Django 1.3
    from django.db.transaction import state
except ImportError:
    state = None
from django.dispatch import Signal


class TransactionSignals(object):
    """A container for the transaction signals."""

    def __init__(self):
        self.post_commit = Signal()
        self.post_rollback = Signal()
        self.post_transaction_management = Signal()


# Access as django.db.transaction.signals.
transaction.signals = TransactionSignals()


def commit(old_function, *args, **kwargs):
    # This will raise an exception if the commit fails. django.db.transaction
    # decorators catch this and call rollback(), but the middleware doesn't.
    old_function(*args, **kwargs)
    transaction.signals.post_commit.send(None)


def commit_unless_managed(old_function, *args, **kwargs):
    old_function(*args, **kwargs)
    if not transaction.is_managed():
        transaction.signals.post_commit.send(None)


# commit() isn't called at the end of a transaction management block if there
# were no changes. This function is always called so the signal is always sent.
def leave_transaction_management(old_function, *args, **kwargs):
    # If the transaction is dirty, it is rolled back and an exception is
    # raised. We need to send the rollback signal before that happens.
    if transaction.is_dirty():
        transaction.signals.post_rollback.send(None)

    old_function(*args, **kwargs)
    transaction.signals.post_transaction_management.send(None)


def managed(old_function, *args, **kwargs):
    # Turning transaction management off causes the current transaction to be
    # committed if it's dirty. We must send the signal after the actual commit.
    flag = kwargs.get('flag', args[0] if args else None)
    if state is not None:
        using = kwargs.get('using', args[1] if len(args) > 1 else None)
        # Do not commit too early for prior versions of Django 1.3
        thread_ident = thread.get_ident()
        top = state.get(thread_ident, {}).get(using, None)
        commit = top and not flag and transaction.is_dirty()
    else:
        commit = not flag and transaction.is_dirty()
    old_function(*args, **kwargs)

    if commit:
        transaction.signals.post_commit.send(None)


def rollback(old_function, *args, **kwargs):
    old_function(*args, **kwargs)
    transaction.signals.post_rollback.send(None)


def rollback_unless_managed(old_function, *args, **kwargs):
    old_function(*args, **kwargs)
    if not transaction.is_managed():
        transaction.signals.post_rollback.send(None)


# Duck punching!
functions = (
    commit,
    commit_unless_managed,
    leave_transaction_management,
    managed,
    rollback,
    rollback_unless_managed,
)

for function in functions:
    name = function.__name__
    function = partial(function, getattr(transaction, name))
    setattr(transaction, name, function)



# coding=utf-8
import os
import sys

from setuptools import setup, Command, find_packages


class RunTests(Command):
    """RunTests class borrowed from django-celery project
    """
    description = 'Run the django test suite from the tests dir.'
    user_options = []
    extra_args = []

    def run(self):
        from django.core.management import execute_from_command_line
        settings_module_name = 'tests.settings'
        os.environ['DJANGO_SETTINGS_MODULE'] = os.environ.get(
                                                    'DJANGO_SETTINGS_MODULE',
                                                    settings_module_name)
        prev_argv = sys.argv[:]
        try:
            sys.argv = [__file__, 'test'] + self.extra_args
            execute_from_command_line(argv=sys.argv)
        finally:
            sys.argv[:] = prev_argv

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

setup(
    name="django-celery-transactions2",
    version="0.1.6",
    description="Django transaction support for Celery tasks.",
    long_description="See https://github.com/chrisdoble/django-celery-transactions",
    author="Chris Doble",
    author_email="chris@chrisdoble.com",
    url="https://github.com/chrisdoble/django-celery-transactions",
    license="Simplified BSD",
    packages=["djcelery_transactions"],
    tests_require=[
        "celery>=2.2.7",
        "Django>=1.2.4",
        "django-celery>=2.2.7",
    ],
    classifiers=[
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Database",
    ],
    cmdclass={'test': RunTests},
)
